import numpy as np

from .atlas import atlas_for


def numeric_jacobian_colwise(f, x, atlas=None, output_atlas=None, delta=1e-8):
    """
    Compute the jacobian of f at x using finite differences.
    See the documentation for `numeric_jacobian` for details. This function
    differs only in that it is a generator that yields each column of the numeric
    jacobian in turn.
    """
    y = f(x)
    input_atlas = atlas_for(x, atlas)
    output_atlas =  atlas_for(y, output_atlas)
    xdof = input_atlas.dof(x)
    ydof = output_atlas.dof(y)
    for i in range(xdof):
        vi = (np.arange(xdof) == i).astype(int)
        yield output_atlas.displacement(f(input_atlas.perturb(x, -vi*delta)),
                                        f(input_atlas.perturb(x, vi*delta))) / (2. * delta)


def numeric_jacobian(f, x, atlas=None, output_atlas=None, delta=1e-8):
    """
    Compute the jacobian for f at x using finite differences.
    """
    return np.array(list(numeric_jacobian_colwise(f, x, atlas, output_atlas, delta))).T



import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal

import atlas


class Foo(object):
	Atlas = "abc"

class AtlasTest(unittest.TestCase):
	def test_atlas_for(self):
		f = Foo()
		self.assertEqual(atlas.atlas_for(f), "abc")
		self.assertIs(atlas.atlas_for(1), atlas.ScalarAtlas)
		self.assertIs(atlas.atlas_for(np.arange(3)), atlas.CartesianAtlas)

	def test_cartesian_atlas(self):
		a = np.array([1, 2, 3])
		assert_array_almost_equal(atlas.CartesianAtlas.perturb(a, [1, 0, -1]), [2, 2, 2])
		assert_array_almost_equal(atlas.CartesianAtlas.displacement(a, [2, 2, 2]), [1, 0, -1])

	def test_scalar_atlas(self):
		self.assertEqual(atlas.ScalarAtlas.perturb(3, [-5]), -2)
		assert_array_almost_equal(atlas.ScalarAtlas.displacement(2, 6), [4])


if __name__ == '__main__':
	unittest.main()



import numpy as np


def atlas_for(v, atlas=None):
    """
    Get the atlas for the object v.
    First, if the atlas keyword parameter is not None then it is returned.
    Next, type(v).Atlas is defined then it is returned.
    Next, if v is a scalar then ScalarAtlas is returned.
    Otherwise, CartesianAtlas is returned.
    """
    if atlas is None:
        try:
            return type(v).Atlas
        except AttributeError:
            if np.isscalar(v):
                return ScalarAtlas
            else:
                return CartesianAtlas
    else:
        return atlas


class ScalarAtlas(object):
    """
    Represents an atlas for real numbers.
    """
    @classmethod
    def dof(cls, v):
        return 1

    @classmethod
    def perturb(cls, v, tangent):
        return v + tangent[0]

    @classmethod
    def displacement(cls, v1, v2):
        """Get a vector v such that perturb(v1, v) = v2."""
        return np.array([v2 - v1])


class CartesianAtlas(object):
    """
    Represents an atlas for finite-dimensional vector spaces.
    """
    @classmethod
    def dof(cls, v):
        return len(v)

    @classmethod
    def perturb(cls, v, tangent):
        return (v + tangent).copy()

    @classmethod
    def displacement(cls, v1, v2):
        """Get a vector v such that perturb(v1, v) = v2."""
        return (v2 - v1).copy()


class HomogeneousAtlas(object):
    """
    Represents an atlas for finite-dimensional projective spaces.
    """
    @classmethod
    def dof(cls, v):
        return len(v) - 1

    @classmethod
    def perturb(cls, v, delta):
        return np.r_[v[:-1] + delta, v[-1]]

    @classmethod
    def displacement(cls, v1, v2):
        v1, w1 = v1[:-1], v1[-1]
        v2, w2 = v2[:-1], v2[-1]
        return (w1 * v2 - w2 * v1) / w2



import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal

from finite_differences import numeric_jacobian, numeric_jacobian_colwise


class Foo(object):
	def __init__(self, x):
		self.x = x

	class Atlas(object):
		@classmethod
		def dof(cls, x):
			return 1

		@classmethod
		def perturb(cls, f, v):
			return Foo(f.x + v[0])

		@classmethod
		def displacement(cls, f1, f2):
			return np.array([f2.x - f1.x])


class FiniteDifferencesTest(unittest.TestCase):
	def test_numeric_jacobian_scalar(self):
		f = lambda x: x**2
		assert_array_almost_equal(numeric_jacobian(f, 3.), [[6.]])

	def test_numeric_jacobian_vector_output(self):
		f = lambda x: np.array([x**2, 3.-x])
		assert_array_almost_equal(numeric_jacobian(f, 3.), [[6.], [-1.]])

	def test_numeric_jacobian_vector_input(self):
		f = lambda x: x[0] * 2.
		assert_array_almost_equal(numeric_jacobian(f, [3., 4.]), [[2., 0.]])

	def test_numeric_jacobian_vector_inout(self):
		f = lambda x: np.array((x[0] * 2., x[1]**2))
		assert_array_almost_equal(numeric_jacobian(f, [3., 4.]), [[2., 0.], [0., 8.]])

	def test_numeric_jacobian_custom_input(self):
		f = lambda f: f.x ** 2
		assert_array_almost_equal(numeric_jacobian(f, Foo(3.)), [[6.]])

	def test_numeric_jacobian_custom_output(self):
		f = lambda x: Foo(x ** 2)
		assert_array_almost_equal(numeric_jacobian(f, 3.), [[6.]])

	def test_numeric_jacobian_custom_inou(self):
		f = lambda f: Foo(f.x ** 2)
		assert_array_almost_equal(numeric_jacobian(f, Foo(3.)), [[6.]])


if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python
from distutils.core import setup

setup(name='manifolds',
      description='Utilities for working with differentiable manifolds',
      version='0.11',
      author='Alex Flint',
      author_email='alex.flint@gmail.com',
      url='https://github.com/alexflint/manifolds',
      packages=['manifolds'],
      package_dir={'manifolds': '.'},
      )
